@@ -1,6 +1,6 @@
 /********************************************************************************
 * ReactPhysics3D physics library, http://www.reactphysics3d.com                 *
-* Copyright (c) 2010-2020 Daniel Chappuis                                       *
+* Copyright (c) 2010-2024 Daniel Chappuis                                       *
 *********************************************************************************
 *                                                                               *
 * This software is provided 'as-is', without any express or implied warranty.   *
@@ -52,36 +52,19 @@ class Deque {
 
     private:
 
-        // -------------------- Constants -------------------- //
-
-        /// Number of items in a chunk
-        const uint CHUNK_NB_ITEMS = 17;
-
-        /// First item index in a chunk
-        const uint CHUNK_FIRST_ITEM_INDEX = CHUNK_NB_ITEMS / 2;
-
         // -------------------- Attributes -------------------- //
 
-        /// Array of chunks
-        T** mChunks;
+        /// Buffer of elements
+        T* mBuffer;
 
         /// Number of current elements in the deque
-        size_t mSize;
+        uint64 mSize;
 
-        /// Number of chunks
-        size_t mNbChunks;
+        /// Capacity
+        uint64 mCapacity;
 
-        /// Index of the chunk with the first element of the deque
-        size_t mFirstChunkIndex;
-
-        /// Index of the chunk with the last element of the deque
-        size_t mLastChunkIndex;
-
-        /// Index of the first element in the first chunk
-        uint8 mFirstItemIndex;
-
-        /// Index of the last element in the last chunk
-        uint8 mLastItemIndex;
+        /// Index in the buffer of the first item of the deque
+        uint64 mFirstItemIndex;
 
         /// Memory allocator
         MemoryAllocator& mAllocator;
@@ -89,87 +72,52 @@ class Deque {
         // -------------------- Methods -------------------- //
 
         /// Return a reference to an item at the given virtual index in range [0; mSize-1]
-        T& getItem(size_t virtualIndex) const {
+        T& getItem(uint64 index) const {
 
-            // If the virtual index is valid
-            if (virtualIndex < mSize) {
+            // Ensure the virtual index is valid
+            assert(index < mSize);
 
-                size_t chunkIndex = mFirstChunkIndex;
-                size_t itemIndex = mFirstItemIndex;
-
-                const size_t nbItemsFirstChunk = CHUNK_NB_ITEMS - mFirstItemIndex;
-                if (virtualIndex < nbItemsFirstChunk) {
-                   itemIndex += virtualIndex;
-                }
-                else {
-
-                    virtualIndex -= nbItemsFirstChunk;
-                    chunkIndex++;
-
-                    chunkIndex += virtualIndex / CHUNK_NB_ITEMS;
-                    itemIndex = virtualIndex % CHUNK_NB_ITEMS;
-                }
-
-                return mChunks[chunkIndex][itemIndex];
-            }
-            else {
-                assert(false);
-            }
+            return mBuffer[mFirstItemIndex + index];
         }
 
         /// Add more chunks
-        void expandChunks(size_t atLeastNbChunks = 0) {
+        void reserve(uint64 capacity) {
 
-            // If it is not necessary to expand the chunks
-            if (atLeastNbChunks > 0 && atLeastNbChunks <= mNbChunks) {
-                return;
-            }
+            if (capacity <= mCapacity) return;
 
-            size_t newNbChunks = mNbChunks == 0 ? 3 : 2 * mNbChunks - 1;
-            if (atLeastNbChunks > 0 && newNbChunks < atLeastNbChunks) {
-                newNbChunks = size_t(atLeastNbChunks / 2) * 2 + 1;
-            }
-            const size_t halfNbChunksToAdd = mNbChunks == 0 ? 1 : (mNbChunks - 1) / 2;
+            // Make sure capacity is an integral multiple of alignment
+            capacity = std::ceil(capacity / float(GLOBAL_ALIGNMENT)) * GLOBAL_ALIGNMENT;
 
-            // Allocate memory for the new array of pointers to chunk
-            void* newMemory = mAllocator.allocate(newNbChunks * sizeof(T*));
-            assert(newMemory != nullptr);
-            T** newChunks = static_cast<T**>(newMemory);
+            // Allocate memory for the new array
+            void* newMemory = mAllocator.allocate(capacity * sizeof(T));
+            T* destination = static_cast<T*>(newMemory);
 
-            // If chunks have already been allocated
-            if (mNbChunks > 0) {
+            const uint64 newStartIndex = capacity / 2 - 1;
 
-                // Copy the pointers to the previous chunks to the new allocated memory location
-                std::uninitialized_copy(mChunks, mChunks + mNbChunks, newChunks + halfNbChunksToAdd);
+            if (mBuffer != nullptr) {
+
+                if (mSize > 0) {
+
+                    // Copy the elements to the new allocated memory location
+                    std::uninitialized_copy(mBuffer + mFirstItemIndex, mBuffer + mFirstItemIndex + mSize, destination + newStartIndex);
+
+                    // Destruct the previous items
+                    for (uint64 i=0; i < mSize; i++) {
+                        mBuffer[mFirstItemIndex + i].~T();
+                    }
+                }
 
                 // Release the previously allocated memory
-                mAllocator.release(mChunks, mNbChunks * sizeof(T*));
+                mAllocator.release(mBuffer, mCapacity * sizeof(T));
             }
 
-            mChunks = newChunks;
+            mBuffer = destination;
+            assert(mBuffer != nullptr);
 
-            // If we need to allocate the first chunk (in the middle of the chunks array)
-            if (mNbChunks == 0) {
-                mChunks[newNbChunks / 2] = static_cast<T*>(mAllocator.allocate(sizeof(T) * CHUNK_NB_ITEMS));
-                assert(mChunks[newNbChunks / 2] != nullptr);
-            }
-
-            mNbChunks = newNbChunks;
-
-            // Allocate memory for each new chunk
-            for (size_t i=0; i < halfNbChunksToAdd; i++) {
-
-                // Allocate memory for the new chunk
-                mChunks[i] = static_cast<T*>(mAllocator.allocate(sizeof(T) * CHUNK_NB_ITEMS));
-                assert(mChunks[i] != nullptr);
-
-                mChunks[mNbChunks - 1 - i] = static_cast<T*>(mAllocator.allocate(sizeof(T) * CHUNK_NB_ITEMS));
-                assert(mChunks[mNbChunks - 1 -i] != nullptr);
-            }
+            mCapacity = capacity;
 
             // Update the first and last chunk index
-            mFirstChunkIndex += halfNbChunksToAdd;
-            mLastChunkIndex += halfNbChunksToAdd;
+            mFirstItemIndex = newStartIndex;
         }
 
     public:
@@ -182,7 +130,7 @@ class Deque {
 
             private:
 
-                size_t mVirtualIndex;
+                uint64 mVirtualIndex;
                 const Deque<T>* mDeque;
 
             public:
@@ -197,15 +145,7 @@ class Deque {
                 using iterator_category = std::random_access_iterator_tag;
 
                 /// Constructor
-                Iterator() = default;
-
-                /// Constructor
-                Iterator(const Deque<T>* deque, size_t virtualIndex) : mVirtualIndex(virtualIndex), mDeque(deque) {
-
-                }
-
-                /// Copy constructor
-                Iterator(const Iterator& it) : mVirtualIndex(it.mVirtualIndex), mDeque(it.mDeque) {
+                Iterator(const Deque<T>* deque, uint64 index) : mVirtualIndex(index), mDeque(deque) {
 
                 }
 
@@ -227,29 +167,29 @@ class Deque {
                     return &(mDeque->getItem(mVirtualIndex));
                 }
 
-                /// Post increment (it++)
+                /// Pre increment (++it)
                 Iterator& operator++() {
                     assert(mVirtualIndex < mDeque->mSize);
                     mVirtualIndex++;
                     return *this;
                 }
 
-                /// Pre increment (++it)
-                Iterator operator++(int number) {
+                /// Post increment (it++)
+                Iterator operator++(int /*number*/) {
                     assert(mVirtualIndex < mDeque->mSize);
                     Iterator tmp = *this;
                     mVirtualIndex++;
                     return tmp;
                 }
 
-                /// Post decrement (it--)
+                /// Pre decrement (--it)
                 Iterator& operator--() {
                     mVirtualIndex--;
                     return *this;
                 }
 
-                /// Pre decrement (--it)
-                Iterator operator--(int number) {
+                /// Post decrement (it--)
+                Iterator operator--(int /*number*/) {
                     Iterator tmp = *this;
                     mVirtualIndex--;
                     return tmp;
@@ -326,35 +266,28 @@ class Deque {
 
         /// Constructor
         Deque(MemoryAllocator& allocator)
-            : mChunks(nullptr), mSize(0), mNbChunks(0), mFirstChunkIndex(1),
-              mLastChunkIndex(1), mFirstItemIndex(CHUNK_FIRST_ITEM_INDEX),
-              mLastItemIndex(CHUNK_FIRST_ITEM_INDEX), mAllocator(allocator) {
+            : mBuffer(nullptr), mSize(0), mCapacity(0), mFirstItemIndex(0), mAllocator(allocator) {
 
-            // Allocate memory for the chunks array
-            expandChunks();
         }
 
         /// Copy constructor
         Deque(const Deque<T>& deque)
-            : mSize(0), mNbChunks(0), mFirstChunkIndex(1),
-              mLastChunkIndex(1), mFirstItemIndex(CHUNK_FIRST_ITEM_INDEX),
-              mLastItemIndex(CHUNK_FIRST_ITEM_INDEX), mAllocator(deque.mAllocator) {
+            : mBuffer(nullptr), mSize(0), mCapacity(0), mFirstItemIndex(deque.mFirstItemIndex),
+              mAllocator(deque.mAllocator) {
 
-            // Allocate memory for the array of chunks
-            expandChunks(deque.mNbChunks);
+            // Allocate memory
+            reserve(deque.mCapacity);
 
             if (deque.mSize > 0) {
 
-                const size_t dequeHalfSize1 = std::ceil(deque.mSize / 2.0f);
-                const size_t dequeHalfSize2 = deque.mSize - dequeHalfSize1;
-
                 // Add the items into the deque
-                for(size_t i=0; i < dequeHalfSize1; i++) {
-                   addFront(deque[dequeHalfSize1 - 1 - i]);
-                }
-                for(size_t i=0; i < dequeHalfSize2; i++) {
-                   addBack(deque[dequeHalfSize1 + i]);
+                for(uint64 i=0; i < deque.mSize; i++) {
+
+                    // Construct the element at its location in the buffer
+                    new (static_cast<void*>(&(mBuffer[deque.mFirstItemIndex + i]))) T(deque.mBuffer[deque.mFirstItemIndex + i]);
                 }
+
+                mSize = deque.mSize;
             }
         }
 
@@ -363,76 +296,58 @@ class Deque {
 
             clear();
 
-            // Release each chunk
-            for (size_t i=0; i < mNbChunks; i++) {
+            if (mCapacity > 0) {
 
-                mAllocator.release(mChunks[i], sizeof(T) * CHUNK_NB_ITEMS);
+                assert(mBuffer != nullptr);
+
+                // Release the chunks array
+                mAllocator.release(mBuffer, sizeof(T) * mCapacity);
             }
 
-            // Release the chunks array
-            mAllocator.release(mChunks, sizeof(T*) * mNbChunks);
+            mCapacity = 0;
+            mBuffer = nullptr;
         }
 
         /// Add an element at the end of the deque
         void addBack(const T& element) {
 
             // If we need to add the item in a another chunk
-            if (mLastItemIndex == CHUNK_NB_ITEMS - 1) {
+            if (mFirstItemIndex + mSize >= mCapacity) {
 
-                // If we need to add more chunks
-                if (mLastChunkIndex == mNbChunks - 1) {
-
-                    // Add more chunks
-                    expandChunks();
-                }
-
-                mLastItemIndex = 0;
-                mLastChunkIndex++;
-            }
-            else if (mSize != 0) {
-                mLastItemIndex++;
+                reserve(mCapacity == 0 ? GLOBAL_ALIGNMENT : mCapacity * 2);
             }
 
+            assert(mFirstItemIndex + mSize < mCapacity);
+
             // Construct the element at its location in the chunk
-            new (static_cast<void*>(&(mChunks[mLastChunkIndex][mLastItemIndex]))) T(element);
+            new (static_cast<void*>(&(mBuffer[mFirstItemIndex + mSize]))) T(element);
 
             mSize++;
 
-            assert(mFirstChunkIndex >= 0 && mLastChunkIndex < mNbChunks);
-            assert(mFirstItemIndex >= 0 && mFirstItemIndex < CHUNK_NB_ITEMS);
-            assert(mLastItemIndex >= 0 && mLastItemIndex < CHUNK_NB_ITEMS);
-            assert(mFirstChunkIndex <= mLastChunkIndex);
+            assert(mFirstItemIndex + mSize <= mCapacity);
+            assert(mSize <= mCapacity);
         }
 
         /// Add an element at the front of the deque
         void addFront(const T& element) {
 
-            // If we need to add the item in another chunk
+            // If we need to add the item in a another chunk
             if (mFirstItemIndex == 0) {
 
-                // If we need to add more chunks
-                if (mFirstChunkIndex == 0) {
-
-                    // Add more chunks
-                    expandChunks();
-                }
-
-                mFirstItemIndex = CHUNK_NB_ITEMS - 1;
-                mFirstChunkIndex--;
-            }
-            else if (mSize != 0) {
-                mFirstItemIndex--;
+                reserve(mCapacity == 0 ? GLOBAL_ALIGNMENT : mCapacity * 2);
             }
 
+            assert(mFirstItemIndex > 0);
+
+            mFirstItemIndex--;
+
             // Construct the element at its location in the chunk
-            new (static_cast<void*>(&(mChunks[mFirstChunkIndex][mFirstItemIndex]))) T(element);
+            new (static_cast<void*>(&(mBuffer[mFirstItemIndex]))) T(element);
 
             mSize++;
 
-            assert(mFirstChunkIndex >= 0 && mLastChunkIndex < mNbChunks);
-            assert(mFirstItemIndex >= 0 && mFirstItemIndex < CHUNK_NB_ITEMS);
-            assert(mLastItemIndex >= 0 && mLastItemIndex < CHUNK_NB_ITEMS);
-            assert(mFirstChunkIndex <= mLastChunkIndex);
+            assert(mFirstItemIndex + mSize <= mCapacity);
+            assert(mSize <= mCapacity);
         }
 
         /// Remove the first element of the deque
@@ -441,28 +356,17 @@ class Deque {
             if (mSize > 0) {
 
                 // Call the destructor of the first element
-                mChunks[mFirstChunkIndex][mFirstItemIndex].~T();
+                mBuffer[mFirstItemIndex].~T();
 
                 mSize--;
+                mFirstItemIndex++;
 
                 if (mSize == 0) {
-                    mFirstChunkIndex = mNbChunks / 2;
-                    mFirstItemIndex = CHUNK_FIRST_ITEM_INDEX;
-                    mLastChunkIndex = mFirstChunkIndex;
-                    mLastItemIndex = CHUNK_FIRST_ITEM_INDEX;
-                }
-                else if (mFirstItemIndex == CHUNK_NB_ITEMS - 1){
-                    mFirstChunkIndex++;
-                    mFirstItemIndex = 0;
-                }
-                else {
-                    mFirstItemIndex++;
+                    mFirstItemIndex = mCapacity / 2 - 1;
                 }
 
-                assert(mFirstChunkIndex >= 0 && mLastChunkIndex < mNbChunks);
-                assert(mFirstItemIndex >= 0 && mFirstItemIndex < CHUNK_NB_ITEMS);
-                assert(mLastItemIndex >= 0 && mLastItemIndex < CHUNK_NB_ITEMS);
-                assert(mFirstChunkIndex <= mLastChunkIndex);
+                assert(mFirstItemIndex + mSize <= mCapacity);
+                assert(mSize <= mCapacity);
             }
         }
 
@@ -472,45 +376,29 @@ class Deque {
             if (mSize > 0) {
 
                 // Call the destructor of the last element
-                mChunks[mLastChunkIndex][mLastItemIndex].~T();
+                mBuffer[mFirstItemIndex + mSize - 1].~T();
 
                 mSize--;
 
                 if (mSize == 0) {
-                    mFirstChunkIndex = mNbChunks / 2;
-                    mFirstItemIndex = CHUNK_FIRST_ITEM_INDEX;
-                    mLastChunkIndex = mFirstChunkIndex;
-                    mLastItemIndex = CHUNK_FIRST_ITEM_INDEX;
-                }
-                else if (mLastItemIndex == 0){
-                    mLastChunkIndex--;
-                    mLastItemIndex = CHUNK_NB_ITEMS - 1;
-                }
-                else {
-                    mLastItemIndex--;
+                    mFirstItemIndex = mCapacity / 2 - 1;
                 }
 
-                assert(mFirstChunkIndex >= 0 && mLastChunkIndex < mNbChunks);
-                assert(mFirstItemIndex >= 0 && mFirstItemIndex < CHUNK_NB_ITEMS);
-                assert(mLastItemIndex >= 0 && mLastItemIndex < CHUNK_NB_ITEMS);
-                assert(mFirstChunkIndex <= mLastChunkIndex);
+                assert(mFirstItemIndex + mSize <= mCapacity);
+                assert(mSize <= mCapacity);
             }
         }
 
         /// Return a reference to the first item of the deque
         const T& getFront() const {
-            if (mSize > 0) {
-                return mChunks[mFirstChunkIndex][mFirstItemIndex];
-            }
-            assert(false);
+            assert(mSize > 0);
+            return mBuffer[mFirstItemIndex];
         }
 
         /// Return a reference to the last item of the deque
         const T& getBack() const {
-            if (mSize > 0) {
-                return mChunks[mLastChunkIndex][mLastItemIndex];
-            }
-            assert(false);
+            assert(mSize > 0);
+            return mBuffer[mFirstItemIndex + mSize - 1];
         }
 
         /// Clear the elements of the deque
@@ -519,32 +407,28 @@ class Deque {
             if (mSize > 0) {
 
                 // Call the destructor of every items
-                for (size_t i=0; i < mSize; i++) {
+                for (uint64 i=0; i < mSize; i++) {
                     getItem(i).~T();
                 }
 
                 mSize = 0;
-
-                mFirstChunkIndex = mNbChunks / 2;
-                mLastChunkIndex = mFirstChunkIndex;
-                mFirstItemIndex = CHUNK_FIRST_ITEM_INDEX;
-                mLastItemIndex = CHUNK_FIRST_ITEM_INDEX;
+                mFirstItemIndex = mCapacity / 2 - 1;
             }
         }
 
         /// Return the number of elements in the deque
-        size_t size() const {
+        uint64 size() const {
             return mSize;
         }
 
         /// Overloaded index operator
-        T& operator[](const uint index) {
+        T& operator[](const uint64 index) {
            assert(index < mSize);
            return getItem(index);
         }
 
         /// Overloaded const index operator
-        const T& operator[](const uint index) const {
+        const T& operator[](const uint64 index) const {
            assert(index < mSize);
            return getItem(index);
         }
@@ -554,7 +438,7 @@ class Deque {
 
             if (mSize != deque.mSize) return false;
 
-            for (size_t i=0; i < mSize; i++) {
+            for (uint64 i=0; i < mSize; i++) {
                 if (getItem(i) != deque.getItem(i)) {
                     return false;
                 }
@@ -577,27 +461,25 @@ class Deque {
                 // Clear all the elements
                 clear();
 
+                reserve(deque.mCapacity);
+
                 if (deque.mSize > 0) {
 
-                    // Number of used chunks
-                    const size_t nbUsedChunks = deque.mLastChunkIndex - deque.mFirstChunkIndex + 1;
-
-                    // Expand the chunk if necessary
-                    expandChunks(nbUsedChunks);
-
-                    const size_t dequeHalfSize1 = std::ceil(deque.mSize / 2.0f);
-                    const size_t dequeHalfSize2 = deque.mSize - dequeHalfSize1;
-
                     // Add the items into the deque
-                    for(size_t i=0; i < dequeHalfSize1; i++) {
-                       addFront(deque[dequeHalfSize1 - 1 - i]);
-                    }
-                    for(size_t i=0; i < dequeHalfSize2; i++) {
-                       addBack(deque[dequeHalfSize1 + i]);
+                    for(uint64 i=0; i < deque.mSize; i++) {
+
+                        // Construct the element at its location in the buffer
+                        new (static_cast<void*>(&(mBuffer[mFirstItemIndex + i]))) T(deque.mBuffer[deque.mFirstItemIndex + i]);
                     }
+
                 }
+
+                mSize = deque.mSize;
             }
 
+            assert(mFirstItemIndex + mSize <= mCapacity);
+            assert(mSize <= mCapacity);
+
             return *this;
         }
 
